# __init__.py
from peakshaving_analyzer.input import Config, load_oeds_config, load_yaml_config
from peakshaving_analyzer.output import Results
from peakshaving_analyzer.PSA import PeakShavingAnalyzer

"""
PeakShaverAnalyzer package initialization.
"""

__all__ = ["PeakShavingAnalyzer", "Config", "Results", "load_yaml_config", "load_oeds_config"]
__version__ = "0.1.1"



import logging
from dataclasses import dataclass
from pathlib import Path

import fine as fn
import pandas as pd
import sqlalchemy

from peakshaving_analyzer import Config
from peakshaving_analyzer.common import IOHandler

log = logging.getLogger(__name__)


@dataclass
class Results(IOHandler):
    # general parameters
    name: str

    # output timeseries
    grid_usage_kw: pd.Series | None = None
    storage_charge_kw: pd.Series | None = None
    storage_discharge_kw: pd.Series | None = None
    storage_soc_kwh: pd.Series | None = None
    solar_generation_kw: pd.Series | None = None
    consumption_kw: pd.Series | None = None
    energy_price_eur: pd.Series | None = None

    # energy costs itself
    energy_costs_eur: float | None = None

    # grid energy and capacity costs
    grid_energy_costs_eur: float | None = None
    grid_capacity_costs_eur: float | None = None
    grid_capacity_kw: float | None = None

    # storage system costs
    storage_invest_eur: float | None = None
    storage_annuity_eur: float | None = None
    storage_capacity_kwh: float | None = None
    inverter_invest_eur: float | None = None
    inverter_annuity_eur: float | None = None
    inverter_capacity_kw: float | None = None

    # solar system costs
    solar_invest_eur: float | None = None
    solar_annuity_eur: float | None = None
    solar_capacity_kwp: float | None = None

    # total costs
    total_yearly_costs_eur: float | None = None
    total_annuity_eur: float | None = None
    total_invest_eur: float | None = None

    def timeseries_to_df(self):
        df = pd.DataFrame()

        df["grid_usage_kw"] = self.grid_usage_kw
        df["storage_charge_kw"] = self.storage_charge_kw
        df["storage_discharge_kw"] = self.storage_discharge_kw
        df["storage_soc_kwh"] = self.storage_soc_kwh
        df["solar_generation_kw"] = self.solar_generation_kw
        df["consumption_kw"] = self.consumption_kw
        df["energy_price_eur"] = self.energy_price_eur

        return df

    def timeseries_to_csv(self, path: str | Path):
        df = self.timeseries_to_df()
        df.to_csv(path)

    def timeseries_to_json(self, path: str | Path):
        df = self.timeseries_to_df()
        df.to_json(path)

    def to_sql(
        self,
        connection: str | sqlalchemy.engine.Connection,
        include_timeseries: bool = True,
        overview_table_name: str = "overview",
        timeseries_table_name: str = "timeseries",
        schema: str = "public",
    ) -> None:
        df = pd.DataFrame([self.to_dict(include_timeseries=False)])
        df.to_sql(name=overview_table_name, schema=schema, con=connection, if_exists="append", index=False)

        if include_timeseries:
            df = self.timeseries_to_df()
            df["name"] = self.name
            df.to_sql(name=timeseries_table_name, schema=schema, con=connection, if_exists="append", index=False)

    def plot_storage_timeseries(self):
        storage_columns = ["storage_charge_kw", "storage_discharge_kw", "storage_soc_kwh"]
        self._plot(cols_to_plot=storage_columns)

    def plot_consumption_timeseries(self):
        consumption_columns = ["grid_usage_kw", "storage_discharge_kw", "solar_generation_kw", "consumption_kw"]
        self._plot(cols_to_plot=consumption_columns)


def create_results(config: Config, esm: fn.EnergySystemModel) -> Results:
    data = {}
    data["name"] = config.name

    _retrieve_timeseries(data, esm=esm, config=config)
    log.info("Retrieved timeseries")

    _retrieve_system_sizes(data, esm=esm)
    log.info("Retrieved system sizes")

    _retrieve_system_costs(data, esm=esm)
    log.info("Retrieved system costs")

    return Results(**data)


def _get_val_from_summary(esm: fn.EnergySystemModel, model_name: str, index: tuple[str], location) -> float:
    try:
        return esm.getOptimizationSummary(model_name).loc[index, location]
    except KeyError:
        log.warning(f"KeyError: {index} not found in {model_name} model.")
        return 0.0


def _get_optimum_ts(esm: fn.EnergySystemModel, model_name: str, variable: str, index: tuple[str]) -> pd.Series:
    """Retrieves optimum timeseries from ESM.

    Args:
        model_name (str): Component model name.
        variable (str): The variable to get.
        index (tuple[str]): Index to use.

    Returns:
        pd.Series: Optimum timeseries
    """

    try:
        mdl = esm.componentModelingDict[model_name]
        vals_df = mdl.getOptimalValues(variable)["values"]

        s = vals_df.loc[index]
    except Exception as e:
        print(model_name, variable, index)
        raise e

    return s


def _retrieve_timeseries(data: dict[str], esm: fn.EnergySystemModel, config: Config) -> None:
    """Writes optimum timeseries to database."""

    data["grid_usage_kw"] = (
        _get_optimum_ts(
            esm=esm,
            model_name="SourceSinkModel",
            variable="operationVariablesOptimum",
            index=("grid", "grid"),
        )
        / config.hours_per_timestep
    )

    if config.add_storage:
        data["storage_charge_kw"] = (
            _get_optimum_ts(
                esm=esm,
                model_name="StorageModel",
                variable="chargeOperationVariablesOptimum",
                index=("storage", "consumption_site"),
            )
            / config.hours_per_timestep
        )

        data["storage_discharge_kw"] = (
            _get_optimum_ts(
                esm=esm,
                model_name="StorageModel",
                variable="dischargeOperationVariablesOptimum",
                index=("storage", "consumption_site"),
            )
            / config.hours_per_timestep
        )

        data["storage_soc_kwh"] = (
            _get_optimum_ts(
                esm=esm,
                model_name="StorageModel",
                variable="stateOfChargeOperationVariablesOptimum",
                index=("storage", "consumption_site"),
            )
            / config.hours_per_timestep
        )

    else:
        data["storage_charge_kw"] = pd.Series(0, index=list(range(config.n_timesteps)))
        data["storage_discharge_kw"] = pd.Series(0, index=list(range(config.n_timesteps)))
        data["storage_soc_kwh"] = pd.Series(0, index=list(range(config.n_timesteps)))

    if config.add_solar:
        data["solar_generation_kw"] = (
            _get_optimum_ts(
                esm=esm,
                model_name="SourceSinkModel",
                variable="operationVariablesOptimum",
                index=("PV", "consumption_site"),
            )
            / config.hours_per_timestep
        )

    else:
        data["solar_generation_kw"] = pd.Series(0, index=list(range(config.n_timesteps)))

    data["consumption_kw"] = config.consumption_timeseries
    data["energy_price_eur"] = config.price_timeseries["grid"]


def _retrieve_system_sizes(data: dict, esm: fn.EnergySystemModel) -> None:
    data["grid_capacity_kw"] = _get_val_from_summary(
        esm=esm,
        model_name="TransmissionModel",
        index=("capacity_price", "capacity", "[kWh]", "grid"),
        location="consumption_site",
    )

    data["storage_capacity_kwh"] = _get_val_from_summary(
        esm=esm,
        model_name="StorageModel",
        index=("storage", "capacity", "[kWh*h]"),
        location="consumption_site",
    )

    data["inverter_capacity_kw"] = _get_val_from_summary(
        esm=esm,
        model_name="ConversionModel",
        index=("from_storage", "capacity", "[kWh]"),
        location="consumption_site",
    )

    data["solar_capacity_kwp"] = _get_val_from_summary(
        esm=esm,
        model_name="SourceSinkModel",
        index=("PV", "capacity", "[kWh]"),
        location="consumption_site",
    )


def _retrieve_system_costs(data: dict[str], esm: fn.EnergySystemModel) -> None:
    # energy itself
    data["energy_costs_eur"] = _get_val_from_summary(
        esm=esm,
        model_name="SourceSinkModel",
        index=("grid", "TAC", "[Euro/a]"),
        location="grid",
    )

    # grid data
    data["grid_energy_costs_eur"] = _get_val_from_summary(
        esm=esm,
        model_name="TransmissionModel",
        index=("capacity_price", "operation", "[kWh*h]", "grid"),
        location="consumption_site",
    )
    data["grid_capacity_costs_eur"] = (
        _get_val_from_summary(
            esm=esm,
            model_name="TransmissionModel",
            index=("capacity_price", "invest", "[Euro]", "grid"),
            location="consumption_site",
        )
        * 2
    )

    # storage data
    data["storage_invest_eur"] = _get_val_from_summary(
        esm=esm,
        model_name="StorageModel",
        index=("storage", "invest", "[Euro]"),
        location="consumption_site",
    )
    data["storage_annuity_eur"] = _get_val_from_summary(
        esm=esm,
        model_name="StorageModel",
        index=("storage", "TAC", "[Euro/a]"),
        location="consumption_site",
    )

    # inverter data
    data["inverter_invest_eur"] = _get_val_from_summary(
        esm=esm,
        model_name="ConversionModel",
        index=("from_storage", "invest", "[Euro]"),
        location="consumption_site",
    )
    data["inverter_annuity_eur"] = _get_val_from_summary(
        esm=esm,
        model_name="ConversionModel",
        index=("from_storage", "TAC", "[Euro/a]"),
        location="consumption_site",
    )

    # solar data
    data["solar_invest_eur"] = _get_val_from_summary(
        esm=esm,
        model_name="SourceSinkModel",
        index=("PV", "invest", "[Euro]"),
        location="consumption_site",
    )
    data["solar_annuity_eur"] = _get_val_from_summary(
        esm=esm,
        model_name="SourceSinkModel",
        index=("PV", "TAC", "[Euro/a]"),
        location="consumption_site",
    )

    # calculate total costs
    data["total_yearly_costs_eur"] = (
        data["energy_costs_eur"]
        + data["grid_energy_costs_eur"]
        + data["grid_capacity_costs_eur"]
        + data["storage_annuity_eur"]
        + data["inverter_annuity_eur"]
        + data["solar_annuity_eur"]
    )
    data["total_annuity_eur"] = data["storage_annuity_eur"] + data["inverter_annuity_eur"] + data["solar_annuity_eur"]
    data["total_invest_eur"] = data["storage_invest_eur"] + data["inverter_invest_eur"] + data["solar_invest_eur"]



import calendar
import logging
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

import pandas as pd
import pgeocode
import requests
import sqlalchemy
import yaml

from peakshaving_analyzer.common import IOHandler

log = logging.getLogger(__name__)


@dataclass
class Config(IOHandler):
    # general parameters
    name: str
    overwrite_existing_optimization: bool = False
    add_storage: bool = True
    add_solar: bool = False
    auto_opt: bool = False
    solver: str | None = "appsi_highs"
    verbose: bool = False

    # timeseries
    consumption_timeseries: pd.Series | None = None
    price_timeseries: pd.Series | None = None
    solar_generation_timeseries: pd.Series | None = None

    # economic parameters
    overwrite_price_timeseries: bool = False
    producer_energy_price: float = 0.1665
    grid_capacity_price: float = 101.22
    grid_energy_price: float = 0.046
    pv_system_lifetime: int = 30
    pv_system_cost_per_kwp: float = 1200.0
    inverter_lifetime: int = 15
    inverter_cost_per_kw: float = 180
    storage_lifetime: int = 15
    storage_cost_per_kwh: float = 0.5
    interest_rate: float = 3

    # technical parameters
    max_storage_size_kwh: float | None = None
    storage_charge_efficiency: float = 0.9
    storage_discharge_efficiency: float = 0.9
    storage_charge_rate: float = 1
    storage_discharge_rate: float = 1
    inverter_efficiency: float = 0.95
    max_pv_system_size_kwp: float | None = None
    pv_system_kwp_per_m2: float = 0.4

    # metadata needed for optimization (set by peakshaving analyzer)
    timestamps: pd.DatetimeIndex | None = None
    n_timesteps: int | None = None
    hours_per_timestep: float | None = None

    def timeseries_to_df(self):
        df = pd.DataFrame()

        df["consumption_kw"] = self.consumption_timeseries
        df["energy_price_eur"] = self.price_timeseries
        df["solar_generation_kw"] = self.solar_generation_timeseries

        return df


def load_yaml_config(config_file_path: Path | str) -> Config:
    config_path = Path(config_file_path)

    # read in configuration file
    with open(config_path) as file:
        data = yaml.safe_load(file)
        log.info("Configuration file loaded")

    # set log level according to flag
    if data.get("verbose", True):
        log.setLevel(level=logging.INFO)
    else:
        log.setLevel(level=logging.WARNING)

    # set config dir var
    data["config_dir"] = config_path.parent

    # read in consumption timeseries
    data["consumption_timeseries"] = pd.read_csv(data["config_dir"] / data["consumption_file_path"])[
        data["consumption_value_column"]
    ]
    log.info("Consumption timeseries loaded")

    # read in timestamps if provided
    if data.get("timestamp_column"):
        data["timestamps"] = pd.read_csv(data["config_dir"] / data["consumption_file_path"])[data["timestamp_column"]]
        log.info("Timestamps loaded")
    else:
        data["timestamps"] = None

    # create metadata for timeseries
    _create_timeseries_metadata(data)
    log.info("Timeseries metadata created")

    # read or create price timeseries
    data["price_timeseries"] = _read_or_create_price_timeseries(data)
    log.info("Price timeseries loaded or created")

    if data["add_solar"]:
        if data["solar_file_path"]:
            data["solar_generation_timeseries"] = pd.read_csv(data["config_dir"] / data["solar_file_path"])[
                data.get("solar_value_column", "value")
            ]
            log.info("Solar generation timeseries loaded")
        elif data["postal_code"]:
            data["solar_generation_timeseries"] = _fetch_solar_timeseries(data)
            log.info("Solar generation timeseries retrieved from brightsky")
        else:
            msg = "No solar generation timeseries available."
            msg += " Setting add_solar to False."
            log.warning(msg)
            data["add_solar"] = False

    _check_timeseries_length(data)

    _remove_unused_keys(data)

    return Config(**data)


def load_oeds_config(
    con: str | sqlalchemy.engine.Connection, profile_id: int, producer_energy_price: float = 0.1665, *args, **kwargs
) -> Config:
    data = {}

    data.update(kwargs)

    # set fixed values
    if "name" not in data:
        data["name"] = "profile_" + str(profile_id)
    data["hours_per_timestep"] = 0.25
    data["producer_energy_price"] = producer_energy_price

    # load consumption timeseries
    data["consumption_timeseries"] = pd.read_sql(
        sql=f"""
        SELECT *
        FROM vea_industrial_load_profiles.load
        WHERE id = {profile_id}
        ORDER BY timestamp ASC
        """,
        con=con,
    )["value"]

    # calculate if consumption is over 2500h full load hours
    is_over_2500h = data["consumption_timeseries"].sum() / 4 / data["consumption_timeseries"].max() > 2500

    if is_over_2500h:
        sql_flh_text = "under"
    else:
        sql_flh_text = "over"

    _create_timeseries_metadata(data)

    data["price_timeseries"] = _read_or_create_price_timeseries(data)

    # get capacity_price
    data["grid_capacity_price"] = pd.read_sql(
        sql=f"""
        SELECT capacity_price_{sql_flh_text}_2500h_eur_per_kw
        FROM vea_industrial_load_profiles.master
        WHERE id = {profile_id}
        """,
        con=con,
    )[f"capacity_price_{sql_flh_text}_2500h_eur_per_kw"].values[0]

    # get energy_price
    data["grid_energy_price"] = pd.read_sql(
        sql=f"""
        SELECT energy_price_{sql_flh_text}_2500h_eur_per_kwh
        FROM vea_industrial_load_profiles.master
        WHERE id = {profile_id}
        """,
        con=con,
    )[f"energy_price_{sql_flh_text}_2500h_eur_per_kwh"].values[0]

    _remove_unused_keys(data)

    return Config(**data)


def _create_timeseries_metadata(data):
    # if no timestamps are given, we create them
    if not data.get("timestamps", None):
        data["n_timesteps"] = len(data["consumption_timeseries"])
        data["leap_year"] = _detect_leap_year(data)
        data["assumed_year"] = _assume_year(data)
        data["timestamps"] = pd.date_range(
            start=f"{data['assumed_year']}-01-01",
            periods=data["n_timesteps"],
            freq=f"{data['hours_per_timestep']}h",
            tz="UTC",
        )
    # otherwise we just create the metadata from the timestamps
    else:
        data["n_timesteps"] = len(data["timestamps"])
        data["leap_year"] = calendar.isleap(data["timestamps"][0].year)
        data["assumed_year"] = data["timestamps"][0].year


def _detect_leap_year(data):
    """
    Detect if given timeseries is a leap year.

    Returns:
        bool: True if the current year is a leap year, False otherwise.
    """

    return data["n_timesteps"] * data["hours_per_timestep"] == 8784


def _assume_year(data):
    """Assumes year for given timeseries.

    Returns:
        int: the assumed year
    """

    log.info("Assuming year.")
    year = datetime.now().year - 1
    if data["leap_year"]:
        while not calendar.isleap(year):
            year -= 1
    else:
        while calendar.isleap(year):
            year -= 1

    log.info(f"Assumed year to be {year}.")

    return year


def _read_or_create_price_timeseries(data):
    # if no filepath is given, we either...
    if not data.get("price_file_path"):
        # throw an error if no price information is given
        if not data.get("producer_energy_price"):
            msg = "No price information found."
            msg += "Please provide either producer_energy_price or "
            msg += "price_file_path in the configuration file."
            log.error(msg)
            raise ValueError

        # ... or create a timeseries from a fixed price
        else:
            return _create_price_timeseries(data)

    # if the filepath is given, we either ...
    else:
        # we overwrite the timeseries by given fixed price
        if data.get("overwrite_price_timeseries"):
            return _create_price_timeseries(data)

        # or just read in the series from file
        else:
            return _read_price_timeseries(data)


def _create_price_timeseries(data):
    """Creates price timeseries from year and given fixed price.

    Returns:
        pd.Series: The price timeseries
    """

    log.info("Creating price timeseries from fixed price.")
    if data["producer_energy_price"] < 0:
        msg = "Producer energy price is negative."
        log.error(msg)
        raise ValueError(msg)

    df = pd.DataFrame()

    year = datetime.now().year - 1
    if data["leap_year"]:
        while not calendar.isleap(year):
            year -= 1
    else:
        while calendar.isleap(year):
            year -= 1

    df["timestamp"] = pd.date_range(
        f"{data['assumed_year']}-01-01",
        freq=f"{data['hours_per_timestep']}h",
        periods=data["n_timesteps"],
    )
    df["grid"] = data["producer_energy_price"]
    df["consumption_site"] = 0

    log.info("Price timeseries successfully created.")

    return df[["grid", "consumption_site"]]


def _read_price_timeseries(data):
    """
    Read the price timeseries from the specified CSV file.

    Returns:
        pd.Series: The price timeseries.
    """
    log.info("Reading price timeseries from CSV file.")
    df = pd.read_csv(data["config_dir"] / data["price_file_path"])
    df.rename(
        columns={data.get("price_value_column", "value"): "grid"},
        inplace=True,
    )
    df["consumption_site"] = 0
    df.loc[df["grid"] < 0, "grid"] = 0  # set negative prices to zero
    log.info("Price timeseries successfully read and processed.")

    return df[["consumption_site", "grid"]]


def _fetch_solar_timeseries(data):
    """
    Read the solar timeseries from brightsky.

    Returns:
        pd.Series: The solar timeseries.
    """
    log.info("Fetching solar timeseries from BrightSky API.")
    # convert postal code to coordinates
    nomi = pgeocode.Nominatim("de")
    q = nomi.query_postal_code(data["postal_code"])
    lat, lon = q["latitude"], q["longitude"]
    log.info(f"Coordinates for postal code {data['postal_code']}: Latitude={lat}, Longitude={lon}")

    # make API Call
    url = f"https://api.brightsky.dev/weather?lat={lat}&lon={lon}&country=DE"
    url += f"&date={data['assumed_year']}-01-01T00:00:00&last_date={data['assumed_year']}-12-31T23:45:00"
    url += "&timezone=auto&format=json"
    log.info(f"Making API call to: {url}")
    weather_data = requests.get(url).json()

    # put data in dataframe
    df = pd.DataFrame(weather_data["weather"])[["solar"]]
    log.info("Solar timeseries data fetched successfully.")

    # rename to location in ESM, add grid column with no operation possible
    df.rename(columns={"solar": "consumption_site"}, inplace=True)
    df["grid"] = 0

    df.fillna(0, inplace=True)

    # resample to match hours per timestep
    if data["hours_per_timestep"] != 1:
        df = _resample_dataframe(df)

    # convert from kWh/m2 to kW
    # kWh/m2/h = kW/m2 = 1000W/m2
    # no converseion necessary, as solar modules are tested with 1000W/m2

    return df


def _resample_dataframe(df: pd.DataFrame, hours_per_timestep: float, assumed_year: int, n_timesteps: int):
    """Resamples given dataframe for provided details.

    Args:
        df (pd.DataFrame): The dataframe to resample.

    Returns:
        pd.DataFrame: the resampled dataframe.
    """

    log.info("Resampling solar timeseries to match your specifications")

    df["timestamp"] = pd.date_range(start=f"{assumed_year}-01-01", periods=len(df), freq="H")

    # upsample
    if hours_per_timestep < 1:
        # set index as needed for upsamling
        df.set_index("timestamp", inplace=True)

        # upsample using forward filling
        df = df.resample(rule=f"{hours_per_timestep}H", origin="start_day").ffill()

        # the last three quarter hours are missing as original timeseries ends on
        # Dec. 12th 23:00 and not 24:00 / Dec. 13th 00:00
        # so we reindex to include the missing timestamps
        df = df.reindex(
            labels=pd.date_range(
                start=f"{assumed_year}",
                periods=n_timesteps,
                freq=f"{hours_per_timestep}H",
            )
        )

        # and fill the newly created timestamps
        df.fillna(method="ffill", inplace=True)

    # downsample
    else:
        # resample
        df = df.resample(rule=f"{hours_per_timestep}H", on="timestamp").mean()

    df.reset_index(drop=True, inplace=True)

    log.info("Successfully resampled solar timeseries.")

    return df


def _check_timeseries_length(data):
    """
    Check if the length of the consumption and price timeseries matches the expected number of timesteps.

    Raises:
        ValueError: If the length of the timeseries does not match.
    """
    log.info("Checking length of timeseries.")
    if len(data["consumption_timeseries"]) != data["n_timesteps"]:
        msg = "Length of consumption timeseries does not match expected number of timesteps. "
        msg += f"Expected number of timesteps: {len(data['n_timesteps'])}, given timesteps: {len(data['consumption_timeseries'])}"
        raise ValueError(msg)
    if len(data["price_timeseries"]) != data["n_timesteps"]:
        msg = "Length of price timeseries does not match expected number of timesteps. "
        msg += f"Expected number of timesteps: {data['n_timesteps']}, given timesteps: {len(data['price_timeseries'])}"
        raise ValueError(msg)
    if "solar_generation_timeseries" in data and len(data["solar_generation_timeseries"]) != data["n_timesteps"]:
        msg = "Length of solar timeseries does not match expected number of timesteps. "
        msg += f"Expected number of timesteps: {data['n_timesteps']}, given timesteps: {len(data['solar_generation_timeseries'])}"
        raise ValueError(msg)
    log.info("Timeseries length check passed.")


def _remove_unused_keys(data):
    """
    Remove unused keys from the data dictionary.
    """
    log.info("Removing unused keys from data.")
    keys_to_remove = [
        "timestamp_column",
        "consumption_file_path",
        "consumption_value_column",
        "price_file_path",
        "price_value_column",
        "solar_file_path",
        "solar_value_column",
        "postal_code",
        "leap_year",
        "assumed_year",
        "config_dir",
    ]
    for key in keys_to_remove:
        data.pop(key, None)
    log.info("Unused keys removed.")



# !/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

import argparse
import sys
from pathlib import Path

import argcomplete


def cli():
    parser = argparse.ArgumentParser(description="Peak Shaving Analyzer CLI")
    parser.add_argument("-c", "--config", type=str, required=True, help="Path to the configuration YAML file")
    parser.add_argument("-o", "--output", type=str, default="std", help="Path to save results (yaml, json or std)")
    parser.add_argument("--solver", type=str, default="appsi_highs", help="Solver to use (optional)")
    parser.add_argument(
        "--output-timeseries",
        type=str,
        default=None,
        help="Path to save resulting timeseries to (not saving if not provided)",
    )

    argcomplete.autocomplete(parser)
    args = parser.parse_args()

    from peakshaving_analyzer.input import load_yaml_config
    from peakshaving_analyzer.PSA import PeakShavingAnalyzer

    # Load config
    config_path = Path(args.config)
    if not config_path.exists():
        print(f"Config file not found: {config_path}", file=sys.stderr)
        sys.exit(1)
    config = load_yaml_config(config_path)

    # Run optimization
    psa = PeakShavingAnalyzer(config=config)
    results = psa.optimize(solver=args.solver)

    print("---------------------------------------------------------")

    # Output results
    if args.output.endswith(".yaml") or args.output.endswith(".yml"):
        results.to_yaml(args.output)
        print(f"Results saved to {args.output} (YAML)")

    elif args.output.endswith(".json"):
        results.to_json(args.output)
        print(f"Results saved to {args.output} (JSON)")
    else:
        results.print(include_timeseries=args.output_timeseries)

    if args.output_timeseries:
        if args.output_timeseries.endswith(".csv"):
            results.timeseries_to_csv(args.output_timeseries)
        elif args.output_timeseries.endswith(".json"):
            results.timeseries_to_json(args.output_timeseries)
        else:
            raise NotImplementedError


if __name__ == "__main__":
    cli()



import json
from dataclasses import asdict, fields
from pathlib import Path

import pandas as pd
import plotly.express as px
import yaml


class IOHandler:
    def print(self, include_timeseries: bool = False):
        for field in fields(self):
            if include_timeseries:
                print(f"{field.name}: {getattr(self, field.name)}")
            elif isinstance(getattr(self, field.name), (pd.Series | pd.DataFrame | pd.DatetimeIndex)):
                continue
            else:
                print(f"{field.name}: {getattr(self, field.name)}")

    def to_dict(self, include_timeseries: bool = True) -> dict:
        if include_timeseries:
            return asdict(self)
        else:
            return {
                k: v
                for k, v in asdict(self).items()
                if not isinstance(v, (pd.Series | pd.DataFrame | pd.DatetimeIndex))
            }

    def to_json(self, path: str | Path):
        with open(path, "w") as f:
            json.dump(self.to_dict(include_timeseries=False), f, indent=4)

    def to_yaml(self, path: str | Path):
        with open(path, "w") as f:
            yaml.safe_dump(self.to_dict(include_timeseries=False), f, sort_keys=False)

    def _plot(self, cols_to_plot: list[str] | None = None):
        ts_df = self.timeseries_to_df()

        if not cols_to_plot:
            cols_to_plot = ts_df.columns.tolist()

        fig = px.line(ts_df, x=ts_df.index, y=cols_to_plot)
        fig.show()

    def plot_timeseries(self):
        self._plot()



import logging

import fine as fn
import numpy as np
import pandas as pd

from peakshaving_analyzer.input import Config
from peakshaving_analyzer.output import Results, create_results

log = logging.getLogger(__name__)


class PeakShavingAnalyzer:
    def __init__(
        self,
        config: Config,
    ) -> None:
        self.config = config

        self.consumption_timeseries = config.consumption_timeseries
        self.price_timeseries = config.price_timeseries
        self.hours_per_timestep = config.hours_per_timestep
        self.number_of_timesteps = len(self.consumption_timeseries)
        self.add_stor = config.add_storage
        self.add_sol = config.add_solar
        self.auto_opt = config.auto_opt
        self.verbose = config.verbose

        self.interest_rate = config.interest_rate

        self.grid_capacity_price = config.grid_capacity_price
        self.grid_energy_price = config.grid_energy_price
        self.producer_energy_price = config.producer_energy_price

        self.storage_cost_per_kwh = config.storage_cost_per_kwh
        self.storage_lifetime = config.storage_lifetime
        self.storage_charge_efficiency = config.storage_charge_efficiency
        self.storage_discharge_efficiency = config.storage_discharge_efficiency
        self.storage_charge_rate = config.storage_charge_rate
        self.storage_discharge_rate = config.storage_discharge_rate
        self.max_storage_size_kwh = config.max_storage_size_kwh
        self.inverter_cost_per_kw = config.inverter_cost_per_kw
        self.inverter_lifetime = config.inverter_lifetime
        self.inverter_efficiency = config.inverter_efficiency

        self.max_pv_system_size_kwp = config.max_pv_system_size_kwp
        self.pv_system_cost_per_kwp = config.pv_system_cost_per_kwp
        self.pv_system_lifetime = config.pv_system_lifetime
        self.pv_system_lifetime = config.pv_system_lifetime

        if self.add_sol:
            self.solar_generation_timeseries = config.solar_generation_timeseries

        self.solver = config.solver

        if config.verbose:
            log.setLevel(level=logging.INFO)
        else:
            log.setLevel(level=logging.WARNING)

        self._create_esm()
        self._add_source()
        self._add_transmission()
        self._add_sink()
        log.info("Built default ESM.")

        if self.add_stor:
            self.add_storage()
            log.info("Added storage.")

        if self.add_sol:
            self.add_solar()
            log.info("Added solar.")

        if self.auto_opt:
            self.optimize(solver=self.solver)
            log.info("Optimized.")

            self.save_results(config)
            log.info("Saved results.")

    def _create_esm(self):
        self.esm = fn.EnergySystemModel(
            locations={"grid", "consumption_site"},
            commodities={"energy", "stored_energy"},
            commodityUnitsDict={"energy": "kWh", "stored_energy": "kWh"},
            costUnit="Euro",
            numberOfTimeSteps=self.number_of_timesteps,
            hoursPerTimeStep=self.hours_per_timestep,
            verboseLogLevel=2,
        )

    def _add_sink(self):
        load_df = pd.DataFrame(
            columns=["grid", "consumption_site"],
            index=np.arange(0, self.number_of_timesteps, 1),
        )

        load_df["grid"] = 0
        load_df["consumption_site"] = self.consumption_timeseries * self.hours_per_timestep

        self.esm.add(
            fn.Sink(
                esM=self.esm,
                commodity="energy",
                name="consumption_site",
                hasCapacityVariable=False,
                operationRateFix=load_df,
            )
        )

    def _add_source(self):
        source_df = pd.DataFrame(
            columns=["grid", "consumption_site"],
            index=np.arange(0, self.number_of_timesteps, 1),
        )

        source_df["grid"] = 1e18
        source_df["consumption_site"] = 0

        self.esm.add(
            fn.Source(
                esM=self.esm,
                commodity="energy",
                name="grid",
                hasCapacityVariable=False,
                operationRateMax=source_df,
                commodityCostTimeSeries=self.price_timeseries,
            )
        )

    def _add_transmission(self):
        self.esm.add(
            fn.Transmission(
                esM=self.esm,
                name="capacity_price",
                commodity="energy",
                hasCapacityVariable=True,
                investPerCapacity=self.grid_capacity_price,
                interestRate=self.interest_rate,
                economicLifetime=1,
                technicalLifetime=1,
            )
        )

    def add_solar(self):
        self.esm.add(
            fn.Source(
                esM=self.esm,
                name="PV",
                commodity="energy",
                hasCapacityVariable=True,
                operationRateMax=self.solar_generation_timeseries,
                capacityMax=self.max_pv_system_size_kwp,
                investPerCapacity=self.pv_system_cost_per_kwp,
                interestRate=self.interest_rate,
                economicLifetime=self.pv_system_lifetime,
                technicalLifetime=self.pv_system_lifetime,
            )
        )

    def add_storage(self):
        self.esm.add(
            fn.Conversion(
                esM=self.esm,
                name="to_storage",
                physicalUnit="kWh",
                commodityConversionFactors={
                    "energy": -1,
                    "stored_energy": self.inverter_efficiency,
                },
                hasCapacityVariable=True,
                investPerCapacity=0,
                linkedConversionCapacityID="storage",
                interestRate=self.interest_rate,
            )
        )

        self.esm.add(
            fn.Storage(
                esM=self.esm,
                name="storage",
                commodity="stored_energy",
                locationalEligibility=pd.Series([1, 0], index=["consumption_site", "grid"]),
                hasCapacityVariable=True,
                cyclicLifetime=10000,
                chargeEfficiency=self.storage_charge_efficiency,
                dischargeEfficiency=self.storage_discharge_efficiency,
                capacityMax=self.max_storage_size_kwh,
                economicLifetime=self.storage_lifetime,
                technicalLifetime=self.storage_lifetime,
                chargeRate=self.storage_charge_rate,
                dischargeRate=self.storage_discharge_rate,
                doPreciseTsaModeling=False,
                investPerCapacity=self.storage_cost_per_kwh,
                interestRate=self.interest_rate,
            )
        )

        self.esm.add(
            fn.Conversion(
                esM=self.esm,
                name="from_storage",
                physicalUnit="kWh",
                commodityConversionFactors={"stored_energy": -1, "energy": 1},
                hasCapacityVariable=True,
                investPerCapacity=self.inverter_cost_per_kw,
                economicLifetime=self.inverter_lifetime,
                technicalLifetime=self.inverter_lifetime,
                linkedConversionCapacityID="storage",
                interestRate=self.interest_rate,
            )
        )

    def optimize(self, solver: str | None = None) -> Results:
        log.info("Creating pyomo model.")
        self.esm.declareOptimizationProblem()

        # add constraint setting storage level on start
        # of optimization to zero
        if self.add_stor:
            self.esm.pyM.stateOfCharge_stor["consumption_site", "storage", 0, 0, 0].setub(0)

        # set solver if not provided
        if not solver:
            solver = self.config.solver

        log.info("Optimizing. Depending on the given parameters and your setup, this may take a while.")

        self.esm.optimize(solver=solver, declaresOptimizationProblem=False)

        results = create_results(self.config, self.esm)

        return results

    def build_and_optimize(
        self,
        add_storage: bool = False,
        storage_cost_per_kwh: float = 145,
        storage_lifetime: int = 15,
        storage_charge_efficiency: float = 0.9,
        storage_discharge_efficiency: float = 0.9,
        storage_charge_rate: float = 1,
        storage_discharge_rate: float = 1,
        max_storage_size_kwh: float | None = None,
        inverter_cost_per_kw: float = 180,
        inverter_lifetime: int = 15,
        inverter_efficiency: float = 0.95,
        add_solar: bool = False,
        solar_data: pd.Series | None = None,
        pv_system_cost_per_kwp: float = 900,
        max_pv_system_size_kwp: float | None = None,
        pv_system_lifetime: int = 30,
    ) -> Results:
        # model building
        self._add_sink()
        self._add_source()
        self._add_transmission()

        if add_storage:
            self.add_storage(
                storage_cost_per_kwh=storage_cost_per_kwh,
                storage_lifetime=storage_lifetime,
                storage_charge_efficiency=storage_charge_efficiency,
                storage_discharge_efficiency=storage_discharge_efficiency,
                storage_charge_rate=storage_charge_rate,
                storage_discharge_rate=storage_discharge_rate,
                max_storage_size_kwh=max_storage_size_kwh,
                inverter_cost_per_kw=inverter_cost_per_kw,
                inverter_lifetime=inverter_lifetime,
                inverter_efficiency=inverter_efficiency,
            )

        if add_solar:
            self.add_solar(
                solar_data=solar_data,
                pv_system_cost_per_kwp=pv_system_cost_per_kwp,
                max_pv_system_size_kwp=max_pv_system_size_kwp,
                pv_system_lifetime=pv_system_lifetime,
            )

        # optimize and return results
        return self.optimize()
